
if (Program.Option("SceneCount").DecimalValue > 0)
      Program.AddVirtualModules("GoogleHome.Scene", "Generic", "homegenie/generic/generic"
                              ,1,(int)Program.Option("SceneCount").DecimalValue);

string jwt_key = "xxxxxxxxxxxxxxxxxxx";

Func<string,string> Base64UrlEncode = (string arg) => {
    string s = Convert.ToBase64String(System.Text.Encoding.Default.GetBytes(arg)); // Regular base64 encoder
    s = s.Split('=')[0]; // Remove any trailing '='s
    s = s.Replace('+', '-'); // 62nd char of encoding
    s = s.Replace('/', '_'); // 63rd char of encoding
    return s;
};

Func<string,string> Base64UrlDecode = (string arg) => {
    string s = arg;
    s = s.Replace('-', '+'); // 62nd char of encoding
    s = s.Replace('_', '/'); // 63rd char of encoding
    switch (s.Length % 4) {
        case 0: break; // No pad chars in this case
        case 2: s += "=="; break; // Two pad chars
        case 3: s += "="; break; // One pad char
        default: throw new System.Exception("Illegal base64url string!");
    }
    return System.Text.Encoding.Default.GetString(Convert.FromBase64String(s)); // Standard base64 decoder
};

Func<string, string> sign = (string msg) => {
	System.Security.Cryptography.HMACSHA256 hmacsha256 = new System.Security.Cryptography.HMACSHA256(System.Text.Encoding.ASCII.GetBytes(jwt_key));
	return hmacsha256.ComputeHash(System.Text.Encoding.ASCII.GetBytes(msg)).ToString();
};

Func<string, dynamic> decode = (string jwt) => {
    string[] tks = jwt.Split('.');
	dynamic header = (dynamic)JsonConvert.DeserializeObject(Base64UrlDecode(tks[0]));
    string sig = Base64UrlDecode(tks[2]);
    if (sig != sign(String.Format("{0}.{1}", tks[0],tks[1]))) {
		throw new System.Exception("Signature verification failed");
    }
  	dynamic payload = (dynamic)JsonConvert.DeserializeObject(Base64UrlDecode(tks[1]));
    return payload;
};

Func<dynamic, string> encode = (dynamic payload) => {
	dynamic header = (dynamic)new ExpandoObject();
	header.typ = "JWT";
	header.alg = "HS256";
    string signing_input = String.Format("{0}.{1}",
		Base64UrlEncode(JsonConvert.SerializeObject(header)),
		Base64UrlEncode(JsonConvert.SerializeObject(payload)));
	return String.Format("{0}.{1}", signing_input,Base64UrlEncode(sign(signing_input)));
};
		
Func<string, string, bool> validToken = (string scope, string Token) => {

	dynamic payload;
	try{
		payload = decode(Token);
    } catch {
      return false;
    }
  
  	if (((string)payload.scope == scope) && (DateTime.Parse((string)payload.exp) > DateTime.Now))
		return true;
    else
    	return false;
};


Func<int, string, string> genToken = (int days, string scope) => {
	dynamic payload = new ExpandoObject();
	payload.sub = "Me";
	payload.iss = "https://";
	payload.aud = "url";
	payload.iat = DateTime.Now.ToString();
	payload.exp = DateTime.Now.AddDays(days).ToString();
	payload.scope = scope;
  	return encode(payload);
};

Func<HttpListenerContext, dynamic, int, bool> processjsonresponse = (HttpListenerContext context, dynamic responsedata, int StatusCode ) => {
    HttpListenerResponse response = context.Response;
  	response.StatusCode = StatusCode;
    Program.Notify("Google Assistant JSON", JsonConvert.SerializeObject(responsedata));
    byte[] buffer = System.Text.Encoding.UTF8.GetBytes(JsonConvert.SerializeObject(responsedata));
    response.ContentLength64 = buffer.Length;
  	response.ContentType = "application/json";
    System.IO.Stream output = response.OutputStream;
    output.Write(buffer,0,buffer.Length);
    output.Close();
  	return true;
};

Func<HttpListenerContext, dynamic, int, bool> processstringresponse = (HttpListenerContext context, dynamic responsedata, int StatusCode ) => {
	HttpListenerResponse response = context.Response;
	response.StatusCode = StatusCode;
	byte[] buffer = System.Text.Encoding.UTF8.GetBytes(responsedata);
    response.ContentLength64 = buffer.Length;
  	response.ContentType = "text/html";
    System.IO.Stream output = response.OutputStream;
    output.Write(buffer,0,buffer.Length);
    output.Close();
  	return true;
};


Func<HttpListenerContext, bool> processAuth = (HttpListenerContext context) => {  
  	bool res = false;
  
    Program.Notify("Google Assistant Auth", context.Request.QueryString["client_id"]);
    if (context.Request.QueryString["redirect_uri"] == Program.Option("RedirectUri").Value
    && context.Request.QueryString["client_id"] == Program.Option("ClientId").Value) {
 	   string url = context.Request.QueryString["redirect_uri"] +
      		"?code=" + genToken(1,"code") +
      		"&state="+ context.Request.QueryString["state"] ;
		string rep = "<html><body><a href=\"" + url + " \"><button>Log In</button></a></body></html>";
 		res = processstringresponse(context, rep, 200);
    } else
	  	res = processstringresponse(context, "unauthorise", 400);
  	return true;
};

Func<HttpListenerContext, string, bool> processToken = (HttpListenerContext context, string requeststring) => {  
	bool res = false;

    Dictionary<string, string> coll = requeststring.Split('&').ToDictionary(item => item.Split('=')[0], item => item.Split('=')[1]);
  
  	dynamic rep = (dynamic)new ExpandoObject();
    if (coll["client_secret"] == Program.Option("ClientSecret").Value
    && coll["client_id"] == Program.Option("ClientId").Value  ) {
		string scope = "";
		string token = "";
	    Program.Notify("Google Assistant Token", coll["grant_type"]);
  		switch (coll["grant_type"]){
		case "authorization_code" :
      		scope = "code";
    	  	token = coll["code"];
			break;
		case "refresh_token" :
      		scope = "refresh";
	      	token = coll["refresh_token"];
			break;
  	  	}  
      	if (validToken(scope, token)) {
			rep.token_type = "bearer";
  			rep.access_token = genToken(1,"execute");
  			rep.refresh_token = genToken(7,"refresh");
  			rep.expires_in = 3600;
 			res = processjsonresponse(context, rep, 200);
        } else {
			rep.error = "invalid_grant";
    	  	res = processjsonresponse(context, rep, 400);
          
        }
    } else {
		rep.error = "invalid_grant";
      	res = processjsonresponse(context, rep, 400);
      
    }
  	return true;
};

Func<dynamic, dynamic, dynamic> processSync = (dynamic req, dynamic rep) => {
	rep.agentUserId = "Me";
  	rep.devices = (dynamic)new List<dynamic>();
    Modules.WithFeature("GoogleAssistant.Link").Each((module) => {
   	  	dynamic device = (dynamic)new ExpandoObject();
		bool haveHue = module.Parameter("Widget.DisplayModule").Value.Equals("homegenie/generic/colorlight");
		rep.devices.Add(device);
 		device.id = module.Instance.Domain + "-" + module.Instance.Address;
      	if (module.IsOfDeviceType("dimmer")){
 			device.type ="action.devices.types.LIGHT";
  			device.traits = (dynamic)new List<dynamic>();
			device.traits.Add("action.devices.traits.OnOff");
 			device.traits.Add("action.devices.traits.Brightness");
            if (haveHue) {
	  			device.traits.Add("action.devices.traits.ColorSpectrum");
	  			device.traits.Add("action.devices.traits.ColorTemperature");
            }
        }
      	if (module.IsOfDeviceType("Sensor")){
 			device.type ="action.devices.types.CAMERA";
  			device.traits = (dynamic)new List<dynamic>();
			device.traits.Add("action.devices.traits.CameraStream");
        }  
      	if (module.IsInDomain("Harmony.Device")){
 			device.type ="action.devices.types.SWITCH";
  			device.traits = (dynamic)new List<dynamic>();
			device.traits.Add("action.devices.traits.OnOff");
 			device.traits.Add("action.devices.traits.Brightness");
        }  
      	if (module.IsInDomain("GoogleHome.Scene")){
 			device.type ="action.devices.types.SCENE";
  			device.traits = (dynamic)new List<dynamic>();
			device.traits.Add("action.devices.traits.Scene");
        }  
      
      
		device.name =  (dynamic)new ExpandoObject();
		device.name.defaultNames = (dynamic)new List<dynamic>();
		device.name.defaultNames.Add(module.Instance.Name);
		device.name.name = module.Instance.Name;
		device.name.nicknames = (dynamic)new List<dynamic>();
		device.name.nicknames.Add(module.Instance.Name);
	   	device.willReportState = false;
      	foreach ( string grp in module.Groups){
            if (module.IsInGroup(grp))
         		device.roomHint = grp;
        }
		device.attributes = (dynamic)new ExpandoObject();
      	if (module.IsOfDeviceType("dimmer")) {
            if (haveHue){
	    		device.attributes.colorModel = "RGB";
	    		device.attributes.temperatureMinK = 2000;
	    		device.attributes.temperatureMaxK = 6500;
            }
        }
      	if (module.IsInDomain("GoogleHome.Scene")){
            device.attributes.sceneReversible = module.HasFeature("GoogleAssistant.Deactivate");
        }

      
		return false; // continue iterating
	});
    return true;
};


Func<double, double, double, int> HSVToRGB = (double H, double S, double V) => {
	double r = 0, g = 0, b = 0;

	if (S == 0)
	{
		r = V;
		g = V;
		b = V;
	}
	else
	{
		int i;
		double f, p, q, t;

		if (H == 360)
			H = 0;
		else
			H = H / 60;

		i = (int)Math.Truncate(H);
		f = H - i;

		p = V * (1.0 - S);
		q = V * (1.0 - (S * f));
		t = V * (1.0 - (S * (1.0 - f)));

		switch (i)
		{
			case 0:
				r = V;
				g = t;
				b = p;
				break;

			case 1:
				r = q;
				g = V;
				b = p;
				break;

			case 2:
				r = p;
				g = V;
				b = t;
				break;

			case 3:
				r = p;
				g = q;
				b = V;
				break;

			case 4:
				r = t;
				g = p;
				b = V;
				break;

			default:
				r = V;
				g = p;
				b = q;
				break;
		}

	}

  	int rr = (int)Math.Round(r * 255);
  	int gg = (int)Math.Round(g * 255);
  	int bb = (int)Math.Round(b * 255);
	Program.Notify("Google Assistant Color", String.Format("{0},{1},{2}",rr,gg,bb));
  	
	return (rr * 0x010000 + gg * 0x0100 + bb);
};

Func<ModuleHelper, dynamic> getState = (ModuleHelper module) => {
		dynamic state = (dynamic)new ExpandoObject();
			state.on = module.IsOn;
			state.online = true;
  			if (module.HasParameter("Status.Level")) {
        	    state.brightness = module.Parameter("Status.Level").DecimalValue * 100;
			}
  			bool haveHue = module.Parameter("Widget.DisplayModule").Value.Equals("homegenie/generic/colorlight");
			if (haveHue) {
	         	string[]hsb = module.Parameter("Status.ColorHsb").Value.Split(',');
    	    	state.color = (dynamic)new ExpandoObject();
        	    double h = Double.Parse(hsb[0]) * 360;
            	double s = Double.Parse(hsb[1]);
	            double v = Double.Parse(hsb[2]);
    	    	state.color.spectrumRGB = HSVToRGB(h,s,v);
        	}
          

  	return state;
};

Func<dynamic, dynamic, dynamic, dynamic> processQuery = (dynamic req, dynamic rep, dynamic payload) => {
  	rep.devices = (dynamic)new ExpandoObject();
	var x = rep.devices as IDictionary<string, dynamic>;
	foreach (dynamic device in payload.devices) {
		string[] id = ((string)(device.id)).Split('-');
		ModulesManager modules = Modules.InDomain(id[0]).WithAddress(id[1]);
        dynamic state = getState(modules.Get());
	    x.Add((string)device.id, state);
	}	
    return true;
};

Func<dynamic, dynamic, dynamic, dynamic> processExecute = (dynamic req, dynamic rep, dynamic payload) => {
  	rep.commands = (dynamic)new List<dynamic>();
	foreach (dynamic command in payload.commands) {
		dynamic repc = (dynamic)new ExpandoObject();
		rep.commands.Add(repc);
		repc.ids = (dynamic)new List<dynamic>();
		repc.status = "SUCCESS";
		foreach (dynamic device in command.devices) {
			repc.ids.Add(device.id);
			string[] id = ((string)(device.id)).Split('-');
			ModulesManager modules = Modules.InDomain(id[0]).WithAddress(id[1]);
			foreach (dynamic execution in command.execution) {
			    Program.Notify("Google Assistant Command", (string)execution.command);
				switch ((string)execution.command){
                  
						case "action.devices.commands.ActivateScene":
							if ((bool)execution["params"].deactivate) 
								modules.Off();
                            else
								modules.On();
                  		break;
                  
						case "action.devices.commands.OnOff":
							if ((bool)execution["params"].on)
							   modules.On();
						    else
							   modules.Off();
						break;
                  
						case "action.devices.commands.BrightnessAbsolute":
                  			int br = execution["params"].brightness;
			            	modules.Command("Control.Level").Set(String.Format("{0}",br));
						break;

                  		case "action.devices.commands.ColorAbsolute":
							if (execution["params"].color.spectrumRGB != null) {
                                int c = execution["params"].color.spectrumRGB;
                              	int r = (c & 0xff0000) / 0x010000;
                              	int g = (c & 0xff00) / 0x0100;
                              	int b = (c & 0xff);
                                modules.Command("Control.Color").Set(String.Format("{0},{1},{2}",r,g,b));
                            } else {
                                int t = execution["params"].color.temperature;
                              	int m = 1000000 / t; 
                                modules.Command("Control.ColorCT").Set(String.Format("{0}",m));
                            }
						break;
				}
					
			}
			repc.states = getState(modules.Get());
		}	
	}
	
  	return true;
};

Func<HttpListenerContext, string, bool> processActionSmartHome = (HttpListenerContext context, string requestString) => {  
	dynamic rep = (dynamic)new ExpandoObject();
 	dynamic req = (dynamic)JsonConvert.DeserializeObject(requestString);
	rep.requestId = req.requestId;
    rep.payload = (dynamic)new ExpandoObject();
	bool res = false;
    int code = 200;
    try {
	    string Token = (string)context.Request.Headers["Authorization"].Split(' ')[1];

    	if (validToken("execute", Token)) {
	
	 	   	foreach (dynamic input in req.inputs) {
			    Program.Notify("Google Assistant Action", (string)input.intent);
    			switch((string)input.intent) {
        		case "action.devices.SYNC" :
        			res = processSync(req, rep.payload);
	        	break;
    	    	case "action.devices.QUERY" :
					res = processQuery(req, rep.payload, input.payload);
        		break;
	        	case "action.devices.EXECUTE" :
					res = processExecute(req, rep.payload, input.payload);
       			break;
				default:
					rep.payload.errorCode = "notSupported";
				break;
    	  		}
    		}
	    } else {
	        Program.Notify("Google Home Action Token", "authFailure");
			rep.payload.errorCode = "authFailure";
    	}
    } catch (Exception e) { 
        Program.Notify("Google Home Action ERROR", e.Message);
		rep.payload.errorCode = "authFailure";
    }  
  	return processjsonresponse(context, rep, code);
};

Func<HttpListenerContext, bool> processRequest = (HttpListenerContext context) => {
   	HttpListenerRequest request = context.Request;
	System.IO.Stream body = request.InputStream;
	System.Text.Encoding encoding = System.Text.Encoding.UTF8;
	System.IO.StreamReader reader = new System.IO.StreamReader(body, encoding);
	string requestString = reader.ReadToEnd();
	body.Close();
	reader.Close();
  			
    Program.Notify("Google Home Url", request.Url.OriginalString);
	if (request.Url.OriginalString.Contains("/google-assistant/auth") ) {
       	processAuth(context);
	} else if (request.Url.OriginalString.Contains("/google-assistant/token")) {
		processToken(context, requestString);
    } else if (request.Url.OriginalString.Contains("/google-assistant")) {
		processActionSmartHome(context, requestString);
    } else {
		dynamic rep = (dynamic)new ExpandoObject();
  		rep.messgae = "error";
  		var res = processjsonresponse(context, rep, 400);
    }
  	return true;
};			

Func<bool> SimpleListener = () =>
{
    // Set up a listener.
    HttpListener listener = new HttpListener();
    listener.Prefixes.Add("http://xxx.xxx.xxx.xxx:7979/");
    listener.Start();
    try {
		while (Program.IsEnabled) {
		    // GetContext blocks while waiting for a request. 
    		HttpListenerContext context = listener.GetContext();
 			processRequest(context);         	
        }
    } finally {
       listener.Stop();
    }
  
    return true;
};

When.WebServiceCallReceived("GoogleHome.Config",( args ) => {
    Program.Notify("Google Assistant Config", "");
    try {
     	// 1 = Address, 2 = Param, 3 = Sub, 4 = value
      	string[] reqs = ((string)args).Split('/');

       	switch (reqs[1]) {
     		case "Sync":
          		string url = String.Format("https://homegraph.googleapis.com/v1/devices:requestSync?key={0}",
                                           Program.Option("ApiKey").Value );
                string body = "{\"agent_user_id\": \"Me\"}";
		    	string rep = Net.WebService(url).Post(body).Call();

    	      	return "{ \"ResponseValue\" : \"OK\"}";
      		default:
		        Program.Notify("Google Home Config ERROR!", (string)args);
				return "{ \"ResponseValue\" : \"ERROR\", \"Message\":\"Invalid action\"}  " + (string)args;
       	}
    } catch (Exception e) { 
        Program.Notify("Google Home Config ERROR!", e.Message);
     	return "{ \"ResponseValue\" : \"ERROR\", \"Message\":\"" + e.Message + "\"}";
    }
}); 


When.WebServiceCallReceived("GoogleHome.Scene", ( args ) => {
    try
    {
        string[] reqs = ((string)args).Split('/');
        var module = Modules.InDomain("GoogleHome.Scene").WithAddress(reqs[1]).Get();

    	string lightnumber = reqs[1];
    	string command = reqs[2];
    	string parameter = ""; 
    	if (reqs.Length > 3) parameter = Uri.UnescapeDataString(reqs[3]);

    	string[] values = parameter.Split(',');
       	var pgm = module.Parameter("GoogleAssistant.Scene").Value;
    	switch(command)
	    {
    		case "Control.On":
               	Net.WebService("http://localhost:8080/api/HomeAutomation.HomeGenie/Automation/Programs.Run/" + pgm).Call();
        	break;
      
			case "Control.Off":
               	Net.WebService("http://localhost:8080/api/HomeAutomation.HomeGenie/Automation/Programs.Break/" + pgm).Call();
        	break;
        
        } 
      
    } catch (Exception e) { 
    
        Program.Notify("GoogleHome.Scene ERROR!", e.Message);
	    return "{ \"ResponseValue\" : \"ERROR\" }";

    }
    return "{ \"ResponseValue\" : \"OK\" }";
});

while (Program.IsEnabled) {
    try {
		SimpleListener();
    } catch (Exception e) { 
        Program.Notify("stop", e.Message);
    }  
  
    Pause(10);
}
